package Catalyst::DB::DbProxy;

# This is a stripped down version of the standard Catalyst::DB::DbProxy
# package.  If you required the extra functionality of that package
# you can just drop it in here.

#############################################################################
#
# Copyright (c) 2005, InternetNZ, The Internet Society of New Zealand
# Incorporated. Wellington, New Zealand.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# Neither the name of the InternetNZ (The Internet Society of New Zealand
# Incorporated) nor the names of its contributors may be used to endorse or
# promote products derived from this software without specific prior written
# permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
#############################################################################


use strict;
use warnings;
use Carp;

use DBI;

sub new {
  my $class = shift;

  my %params = @_;
  my $dbDriverName = $params{dbDriverName}
    or croak "dbDriverName not specified";

  my $dbh_class = "${class}::$dbDriverName";
  eval "require $dbh_class" or die $@;
  return $dbh_class->new(%params);
}

sub init {
  my $self = shift;

  my %params = @_;

  $self->{'dbh'} = DBI->connect($params{'dbConnectString'},
                                $params{'userName'},
                                $params{'password'},
                                { AutoCommit         => $params{'autoCommit'}       || 0,
                                  RaiseError         => $params{'raiseError'}       || 1,
                                  ShowErrorStatement => $params{'showErrorStement'} || 1})
    || croak "Could not connect to db ($params{'dbConnectString'}): $DBI::errstr\n";
}

sub executeRead {
  my $self = shift;
  my %params = @_
    or croak "parameters not supplied to DbProxy::executeRead()";

  my $queryString   = $params{queryString};
  my $queryParams   = $params{queryParams};
  my $compiledQuery = $params{compiledQuery};

  if (! defined $compiledQuery) {
    croak "query not supplied to DB::executeRead()" unless $queryString;

    $compiledQuery = $self->{'dbh'}->prepare($queryString)
      || croak "Failed to prepare db query: $DBI::errstr\n";
  }

  if (!$queryParams) {
    $queryParams = [];
  }
  elsif (!ref($queryParams)) {
    $queryParams = [$queryParams];
  }
  elsif (ref($queryParams) ne 'ARRAY') {
    croak "bad queryParams supplied to DbProxy::executeRead()";
  }

  $compiledQuery->execute(@$queryParams)
    || croak "Failed to execute db query: $DBI::errstr\n";

  return $compiledQuery;
}

sub execute(%) {
    my $self = shift;
    my %params = @_ or croak "parameters not supplied to DbProxy::execute()";

    my $sth = $self->executeRead(@_);
    my $rows = $sth->rows;
    $sth->finish();
    return $rows;
}

sub DESTROY {
    local $@ if $@;  # Don't inadvertently trample caller's $@
    my $self = shift;
    $self->disconnect;
}

sub getAutoCommit($) {
  my $self = shift;
  return $self->{dbh}->{AutoCommit} ? 1 : 0;
}

# Update methods
#
sub rollback() {
    my $self = shift;

    $self->{dbh}->rollback();
}

sub commit() {
  my $self = shift;

  $self->{dbh}->commit();
}

sub disconnect {
    my $self = shift;
    my $result = 1;

    if (my $dbh = $self->{dbh}) {
        eval { $dbh->rollback unless $self->getAutoCommit; };
        $self->{logger}->error('Database rollback failed before disconnect: %s', $@)
            if $@ and $self->{logger};
        $result = $dbh->disconnect;
    }
    delete $self->{dbh};
    return $result;
}

1;
