package PUA;

# ------------------------------------------------------
#   AUTHOR: Andrew Caudwell
#   EMAIL:  andrewc@catalyst.net.nz
#   ------------------------------------------------------
#   APPLICATION: PUA UCI Prototype
#   ------------------------------------------------------
#   FUNCTIONALITY:
#   Personal User Agent used to act on behalf of a
#   Calling/Called Party
#   ------------------------------------------------------

#############################################################################
#
# Copyright (c) 2005, InternetNZ, The Internet Society of New Zealand
# Incorporated. Wellington, New Zealand.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# Neither the name of the InternetNZ (The Internet Society of New Zealand
# Incorporated) nor the names of its contributors may be used to endorse or
# promote products derived from this software without specific prior written
# permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
#############################################################################

=head1 NAME

PUA - Personal User Agent used to act on behalf of a Calling/Called Party

=head1 DESCRIPTION

PUA acts on the behalf of a Calling/Called Party, deciding who can contact
them and by what method at the current time.

=cut

use strict;
use warnings;

use FindBin;
use lib "$FindBin::Bin/../lib/";
use Catalyst::DB::DbProxy::Pg;
use Net::Ping;

#settings
my $PING_TIMEOUT = 2;  # time we will wait to check if a particular host is
                       # tcp pingable before giving up
my $DEBUG = 1;         #toggle debugging

my $log = "$FindBin::Bin/PUA.debug";

#print debug to STDERR if enabled
sub debug {
    my $msg = shift;

    if($DEBUG)
        {
        open  DEBUGLOG, ">>$log" or die("could not open debug log: $!");
        print DEBUGLOG "DEBUG: $msg\n";
        close DEBUGLOG;
        }
}

=head2 new (%args)

Creates a PUA object. Takes a hash containing:

    db       - a connection to the PUA database
    username - the username of the PUA to use

=cut

sub new {
    my ($class,%args) = @_;

    my $self = bless {}, $class;

    $self->{username} = $args{username} || die ("username required");
    $self->{db}       = $args{db}       || die ("db required");

    my $sql = 'select p.pua_id
               from pua p
               where username = ?';

    my $sth = $self->{db}->executeRead(queryString => $sql, queryParams => [$self->{username}]);

    die("No such PUA user '" . $self->{username} . "'\n") unless $sth->rows;

    ($self->{pua_id}) = $sth->fetchrow_array();

    return $self;
}

=head2 device_uri ($device_id, $protocol)

returns the URI for the protocol $protocol of device $device_id.

If no URI found for this criteria, returns undef.

NOTE: Assumes the device will only have one URI for this protocol!

=cut

sub device_uri {
    my ($self, $device_id, $protocol) = @_;

    die("device required")   unless defined $device_id;
    die("protocol required") unless defined $protocol;

    my $db = $self->{db};

    $protocol = "$protocol:%";

    my $sth = $db->executeRead(queryString => 'select uri
                                               from owner_contact_method
                                               where device_id = ? and uri like ?',
                               queryParams => [$device_id,$protocol]);

    return undef unless $sth->rows;

    my($uri) = $sth->fetchrow_array();

    return $uri;
}

#check if inbound uri matches a regex
sub uri_match {
    my($uri, $regex) = @_;

    return $uri =~ /$regex/;
}

#check if a host is reachable (or accepting pings)
sub ping {
    my($host) = @_;

    my $p = new Net::Ping;

    my $result = $p->ping($host, $PING_TIMEOUT);

    $p->close();

    return $result ? 1 : 0;
}

#check current time in seconds is between two times
sub between_times {
    my($start, $finish) = @_;

    debug('start time: '.$start);
    debug('finish time: '.$finish);

    my($sec,$min,$hour) = localtime(time);

    my $comptime = sprintf '%02d%02d', $hour, $min;
    debug('current time: '.$comptime);

    return 1 if($comptime eq $start || $comptime eq $finish);

    if($start lt $finish || $start eq $finish)
      {
  debug('start and finish are same day');
        if($comptime gt $start && $comptime lt $finish)
      {
            debug('current time is between start and finish');
            return 1;
            }
        }
    else
      {
        debug('finish is the following day');
  if($comptime gt $start || $comptime lt $finish)
            {
            debug('current time is between start and finish');
            return 1;
            }
  }

    return 0;
}

=head2 execute_rule($rule,$inbounduri)

given a rule, try the rule and return the resulting hash containing
a device_id and a list of protocols supported by that device.

    device_id
    protocols - list of protocols

This is a recursive function - ie if the rule points to another
rule it, will then execute that rule.

it is ok for a rule to not be accepted, however this method will
die if any rules are found to be invalid (ie refer to
non-existant check rules, etc.

=cut

sub execute_rule {
    my($self, $rule, $inbounduri) = @_;

    my $pua_id = $self->{pua_id};
    my $db     = $self->{db};

    die("rule required")       unless ref $rule eq 'HASH';
    die("inbounduri required") unless defined $inbounduri;

    my $rule_id   = $rule->{rule_id};
    my $rule_name = $rule->{rule_name};
    my $criteria  = $rule->{criteria};
    my @actions   = @{$rule->{action}};

    die("no actions for rule rule_id=$rule_id, rule_name=$rule_name") unless @actions;

    debug("execute_rule $rule_id");

    #check if we meet the criteria
    if(defined $criteria)
      {
  debug("criteria: $criteria");
  #time perdiod criteria - this rule is allowed between some time period
  if($criteria =~ /^hours:(\d\d\d\d)-(\d\d\d\d)$/)
      {
      my ($start, $finish) = ($1,$2);

            return undef unless between_times($start,$finish);
      }
  elsif($criteria =~ /^uri_match:(.+)$/)
            {
            my $regex = $1;
            return undef unless uri_match($inbounduri,$regex);
            }
        elsif($criteria =~ /^ping:(.+)$/)
            {
            my $host = $1;
            return undef unless ping($host);
            }
        elsif ($criteria ne '')
      {
      die("unknown criteria '$criteria' on rule $rule_id");
      }
  }

    foreach my $action (@actions)
      {
  debug("action : $action");

  #connect is the one we're looking for - ie the only action currently that returns a uri
  if($action =~ /^connect:(\d+)$/)
    {
    my $device_id = $1;
    my $sql = 'select uri from owner_contact_method where device_id = ?';
    my $sth = $db->executeRead(queryString => $sql, queryParams => [$device_id]);

    die("no owner_contact_methods for device $device_id") unless $sth->rows;

                my @protocols;

                while(my ($proto) = $sth->fetchrow_array())
                    {
                    #return only the distinct protocols
                    $proto =~ s/^([^:]+):.*$/$1/ or die("invalid uri $proto");
                    push(@protocols, $proto);
        }
                return { device_id => $device_id, protocols => \@protocols };
    }
  #use rule check if allowed, otherwise keep trying other actions on this rule
  elsif($action =~ /^check:(.*)$/)
    {
    my $check_name = $1;
    my $check_rule = $self->get_rule(undef, $check_name);

    die("no such check rule for pua_id=$pua_id, rule_name = $check_name")
      unless defined $check_rule;

    my $check_result = $self->execute_rule($check_rule,$inbounduri);

    return $check_result if defined $check_result;
    }
  else
    {
    die("unknown action ($action)");
    }
  }

    debug("rule not allowed");

    return undef;
}

=head2 get_contacts($uri)

get contacts this uri belongs to

=cut

sub get_contacts {
    my($self, $uri) = @_;

    my $pua_id = $self->{pua_id};
    my $db     = $self->{db};

    die("uri required") unless defined $uri;

    my $sql = "select pua_contact.contact_id
               from contact_method, pua_contact
               where contact_method.contact_id = pua_contact.contact_id
           AND pua_id = ? and uri = ?";

    debug("get_contacts, statement: $sql");
    debug("get_contacts, params: $pua_id, $uri");
    my $sth = $db->executeRead(queryString => $sql, queryParams => [$pua_id, $uri]);

    if (! $sth->rows) {
      debug ("No rows returned");
      return ();
    }

    my @groups = $sth->fetchrow_array();

    while(1)
        {
        $sql = "select parent_contact_id
                from group_entry
                where child_contact_id = ?";

        $sth = $db->executeRead(queryString => $sql, queryParams => [$groups[-1]]);

        last unless $sth->rows;

        push(@groups, $sth->fetchrow_array());
        }

    return @groups;
}

=head2 get_rule($uri,$rule_name)

Select the details of a rule for the pua and optionally one or both of a
$rule_name and a $uri;

The action of the rule is converted to a list split on spaces.

=cut

sub get_rule {
    my($self, $uri, $rule_name) = @_;

    my $pua_id = $self->{pua_id};
    my $db     = $self->{db};

    my $sql = 'select
            r.rule_id,
      r.rule_name,
      r.criteria,
      r.action
      from rule r
      ';

    my @where;
    my @params = ($pua_id);

    debug("rule_name: " . ($rule_name || 'No rule'));
    if(defined $rule_name)
        {
  push(@where, 'r.rule_name = ?');
        push(@params, $rule_name);
  }

    debug("Calling URI: " . ($uri || 'Not present'));
    if(defined $uri)
      {
        my @contacts = $self->get_contacts($uri);
        debug("uri $uri has known contacts: ".join(', ',@contacts));
        return undef unless @contacts;

        push(@where, 'r.contact_id in ('.join(",",('?') x @contacts).')');
  push(@params, @contacts);
  }

    $sql.= 'where
            r.pua_id = ?';

    $sql.=' and '. join(' and ',@where) if @where;

    debug("rule select: \n$sql\n params = ".join(',',@params));

    my $sth = $db->executeRead(queryString => $sql,
                            queryParams => \@params);

    die("multiple rows returned!") if $sth->rows > 1;

    my $rule = $sth->fetchrow_hashref();

    return undef unless(defined $rule);

    # Eat unnecessary whitespace.
    $rule->{action} =~ s/^\s+//;
    $rule->{action} =~ s/\s+$//;
    $rule->{action} =~ s/\s+::://;
    $rule->{action} =~ s/:::\s+//;
    my @action = split(/:::/,delete $rule->{action});
    $rule->{action} = \@action;

    return $rule;
    }

1;
