package PUA::Request;

#############################################################################
#
# Copyright (c) 2005, InternetNZ, The Internet Society of New Zealand
# Incorporated. Wellington, New Zealand.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# Neither the name of the InternetNZ (The Internet Society of New Zealand
# Incorporated) nor the names of its contributors may be used to endorse or
# promote products derived from this software without specific prior written
# permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
#############################################################################

use strict;
use warnings;

use FindBin;
use lib "$FindBin::Bin/../lib/";
use lib "$FindBin::Bin";
use Catalyst::DB::DbProxy::Pg;
use PUA;
use PUA::Transport;
use Data::Dumper;

# turn this off when using sockets since STDERR seems to go to the STDOUT
# of the socket
my $DEBUG = 1;

#database connection parameters
my $DBNAME     = 'pua';
my $DBHOST     = 'localhost';
my $DBUSER     = 'pua';
my $DBPASSWORD = 'pua';

chomp($DBUSER);

my $dbparams = {dbName             => $DBNAME,
                userName           => $DBUSER,
                password           => $DBPASSWORD,
                autoCommit         => 0,
                raiseError         => 1,
                showErrorStatement => 1,
                };

$dbparams->{host} = $DBHOST if(defined($DBHOST) && $DBHOST);

my $log = "$FindBin::Bin/proc_request.debug";

#TODO: more efficient to flush after writing
$| = 1;

#print debug to STDERR if enabled
sub debug {
    my $msg = shift;

    if($DEBUG)
        {
        open  DEBUGLOG, ">>$log" or die("could not open debug log: $!");
        print DEBUGLOG "DEBUG: $msg\n";
        close DEBUGLOG;
        }
}

sub error {
    my $self = shift;

    my $transport = $self->{transport};

    $transport->message('procresponse', error => $@);
    $transport->close();
}

=head2 new

Creates new PUA::Request object. Simply initializes the transport connected to
the provided $stream.

=cut

sub new {
    my ($class,$stream) = @_;

    die("stream required") unless defined $stream;

    my $transport = PUA::Transport->new(stream => $stream);

    my $self = bless { transport => $transport }, $class;

    return $self;
}

=head2 handle

Handles a PUA request on STDIN/OUT. If anything goes wrong, an error message
will be sent to the requester. This method should not die.

=cut

sub handle
    {
    my $self = shift;

    my $transport = $self->{transport};

    eval {
        my $db = new Catalyst::DB::DbProxy::Pg(%$dbparams);

        my $sql;
        my $sth;
        my $puaID;

        my $args = $transport->readRequest();

        debug("data received: ".Dumper($args));

        die("input doesnt contain a procrequest: ".Dumper($args))
            unless defined $args->{procrequest};

        my $procrequest = $args->{procrequest};

        my $username   = $procrequest->{username}   || '';
        my $inboundURI = $procrequest->{inboundURI} || '';

        debug("username = $username , inboundURI = $inboundURI");

        die("username required")   unless $username;
        die("inboundURI required") unless $inboundURI;

        my $pua = PUA->new( db => $db, username => $username );

        my $rule = undef;

        #look for a rule specific to this URI
        if(defined($rule = $pua->get_rule($inboundURI)))
            {
            debug("caller rule: ".$rule->{rule_id});
            }
        #get general case rule
        elsif(defined($rule = $pua->get_rule(undef, 'default')))
            {
            debug("default rule: ".$rule->{rule_id});
            }

        # try rule(s?) until we've found one we can apply or we've exausted
        # all of them.
        my $device = $pua->execute_rule($rule,$inboundURI);

        die("no allowed contact methods found\n") unless defined $device;

        my $device_id = $device->{device_id};
        my $protocols = $device->{protocols};

        die("device_id missing") unless defined $device_id;
        die("protocols missing") unless ref($protocols) eq 'ARRAY' && @$protocols;

        debug("protocols offered for device $device_id : ".join(', ',@$protocols));

        my %response  = (protocols => $protocols);

        debug('response: '.Dumper(\%response));

        $transport->message('procresponse', %response);
        debug('response sent...');

        #wait for protocol to be chosen
        $args = $transport->readRequest();
        debug('read response follow up');

        my $protocol = $args->{procrequest}{protocol}
            or die("client did not respond with a protocol\n");

        debug("client selected protocol: $protocol");

        my $uri = $pua->device_uri($device_id, $protocol)
            or die("could not find uri for device $device_id protocol $protocol");

        %response = (uri => $uri);

        $transport->message('procresponse', %response);
        $transport->close();
        };

    if($@)
        {
        debug("error occured: $@");
        if(defined $transport)
            {
            $self->error($@);
            }
        }

    debug("Finished.");
    }

1;
