package PUA::Transport;

#   ------------------------------------------------------
#   AUTHOR: Andrew Caudwell
#   EMAIL:  andrewc@catalyst.net.nz
#   ------------------------------------------------------
#   APPLICATION: PUA UCI Prototype
#   ------------------------------------------------------
#   FUNCTIONALITY:
#   Allow passing data between pua applications using xml
#   ------------------------------------------------------

#############################################################################
#
# Copyright (c) 2005, InternetNZ, The Internet Society of New Zealand
# Incorporated. Wellington, New Zealand.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# Neither the name of the InternetNZ (The Internet Society of New Zealand
# Incorporated) nor the names of its contributors may be used to endorse or
# promote products derived from this software without specific prior written
# permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
#############################################################################

=head1 NAME

PUA::Transport - Allow passing data between pua applications using xml without
closing streams.

=head1 DESCRIPTION

The PUA::Transport communicates using XML in a specific format. An exmaple
follows

    <puaxmltransaction>
    <puaxml>
       ** XML ENCODED DATA **
    </puaxml>

    ...

    </puaxmltransaction>

The control tags <puaxmltransaction> and <puaxml> must be on their own
seperate lines with no whitespace. Only blank lines are allowed between
control tags (ie the next non empty line after the end of one </puaxml>
should be the next <puaxml> or the </puaxmltransaction>).

=head1 FUNCTIONS

=cut

use Frontier::RPC2;
use IO::Socket;
use Data::Dumper;

use strict;
use warnings;

#TODO: choose a real port, i just pulled this out of the air
my $PUA_PORT = 10667; #;10667;

my $PUAXML_DOCUMENT_NAME = 'puaxmltransaction';
my $PUAXML_ELEMENT_NAME  = 'puaxml';

my $DEBUG = 0;

my $log = "/tmp/transport.debug";

#print debug to STDERR if enabled
sub debug {
    my $msg = shift;

    if($DEBUG)
        {
        open  DEBUGLOG, ">>$log" or die("could not open debug log: $!");
        print DEBUGLOG "DEBUG[$$]: $msg\n";
        close DEBUGLOG;
        }
}

=head2 new ($host, %args)

Connects to the remote PUA of the provided 'host' or 'stream'. If none
provided, it will use STDIN/STDOUT.

=cut

sub new {
    my ($class, %args) = @_;
    my $self =  bless {}, $class;

    my $instream  = undef;
    my $outstream = undef;

    if(defined $args{host})
        {
        #TODO: actually use the uri
        my $socket = new IO::Socket::INET(
                                          PeerAddr => $args{host},
                                          PeerPort => $PUA_PORT,
                                          Proto    => 'tcp',
                                          );
        die "failed to connect to pua:".$args{host}.":$PUA_PORT"
            unless defined $socket;

        #TODO: more efficient to flush only after finished writing
        $socket->autoflush();

        $instream  = $socket;
        $outstream = $socket;
        }
    elsif(defined $args{stream})
        {
        $instream  = $args{stream};
        $outstream = $args{stream};
        }

    $self->{coder} = new Frontier::RPC2;

    $self->{instream} = $instream  || \*STDIN;
    $self->{outstream}= $outstream || \*STDOUT;

    $self->{first_read}      = 1;
    $self->{first_write}     = 1;

    return $self;
}

=head2 encode($method, $data)

encodes a hash into xml using Frontier::RPC2::encode_call($method, $data)
without the XML starting tag.

=cut

sub encode {
    my($self, $method, $data) = @_;

    my $xml = $self->{coder}->encode_call($method, $data);

    #remove the xml header since this isnt a new document
    #TODO: find a better way to do this
    $xml =~ s/^<?xml version="1.0"?>//;

    return $xml;
}

=head2 decode ($rpcxml)

decodes data using Frontier::RPC2::decode and returns a hash in the format:

    { methodname => data }

=cut

sub decode {
    my($self, $xml) = @_;

    my $rpcout = $self->{coder}->decode($xml);

    die("unexpected format of decoded data structure from xml: $xml")
        unless(defined $rpcout->{method_name} && ref($rpcout->{value}) && scalar(@{$rpcout->{value}}) == 1);

    my $out = { $rpcout->{method_name} => $rpcout->{value}[0] };

    return $out;
}

=head2 readRequest()

Reads the next <puaxml> elements off the input stream.

This method will block while waiting for the next EOL character.

=cut

sub readRequest {
    my ($self) = @_;

    debug('called readRequest()');

    my $stream = $self->{instream} or die("input stream required\n");

    my @buff;

    if($self->{first_read})
        {
        debug('looking for document tag');
        my $line = <$stream>;
        die("line is empty!")
            unless defined $line;
        die("first line of stream should be <$PUAXML_DOCUMENT_NAME>: $line")
            unless $line =~ /^<$PUAXML_DOCUMENT_NAME>$/;

  $self->{first_read} = 0;
        }

    my $state = 0;

    debug('waiting for data...');

    while(my $line = <$stream>)
        {
  debug("read line: $line");
        $line =~ s/^\s+//;
  $line =~ s/\s+$//;
        next unless $line;

  if($state==0)
      {
            if($line =~ /^<$PUAXML_ELEMENT_NAME>$/)
        {
    $state = 1;
    next;
    }
            elsif($line =~ /^<\/$PUAXML_DOCUMENT_NAME>$/)
                {
                $self->{first_read} = 1;
                last;
                }
      else
        {
        die("expecting control tag: $line");
        }
      }

  if($line =~ /^<\/$PUAXML_ELEMENT_NAME>$/)
            {
      $state = 0;
      return $self->decode(join("\n",@buff));
            }

  push(@buff, $line);
  }

    die("stream closed without proper end tag </$PUAXML_DOCUMENT_NAME>")
        unless $self->{first_read};

    die("client disconnected cleanly. no request to read.");
}

=head2 message($function, %args)

Writes a message to the output stream inside a <puaxml> tag as a <$function>
element.

If the first request, writes the <puatransaction> document tag.

=cut

sub message {
    my($self, $function, %args) = @_;

    debug("called message: function = '$function', args = ".Dumper(\%args));
    die("function '$function' contains invalid characters")
        if $function =~ /[- ]/;

    my $stream = $self->{outstream}
        or die("output stream required\n");

    my $xml = $self->encode($function,\%args);

    #start a document
    if($self->{first_write})
        {
        print $stream "<$PUAXML_DOCUMENT_NAME>\n";
        $self->{first_write} = 0;
        }

    print $stream "\n<$PUAXML_ELEMENT_NAME>\n";
    print $stream $xml;
    print $stream "\n</$PUAXML_ELEMENT_NAME>\n";
}

=head2 close()

close the transaction by passing a </puatransaction> tag. This doesn't
actually close the stream since it's not ours to close.

=cut

sub close {
    my($self) = @_;

    my $stream = $self->{outstream}
        or die("output stream required\n");

    if($self->{first_write})
        {
        print $stream "<$PUAXML_DOCUMENT_NAME>\n";
        $self->{first_write} = 0;
        }

    $self->{first_write} = 1;

    print $stream "\n</$PUAXML_DOCUMENT_NAME>\n";
}

1;

=head1 AUTHOR

Andrew Caudwell <andrewc@catalyst.net.nz>

=cut
